#!/usr/bin/env python

import sys
import optparse
import collections
import os.path
import os
import subprocess
import multiprocessing
import itertools

import gff
import splicegraph
import fasta
import string
import psg
import table
import chisquare

def sample_results(sample):
    filename = os.path.join(sample, "gene.results.txt")
    return dict([(r["gene_id"], r) for r in table.FileReader(filename)])
    
def chisquare_p(d, df):
    return chisquare.lchisqprob(d, df)

def sample_gene_mle_params(gene, sample_name):
    em_output_filename = os.path.join(sample_name, gene["chrom"],
                                      gene["gene_id"] + ".em.out")
    em_output = psg.parsed_em_output(em_output_filename)
    return em_output["iterations"][-1][1]

def compute_max_delta_psi(gene, sample1_name, sample2_name):
    sample1_params = sample_gene_mle_params(gene, sample1_name)
    sample2_params = sample_gene_mle_params(gene, sample2_name)
    return max([abs(p1 - p2) for p1, p2 in zip(sample1_params, sample2_params)])

def test_diff_splicing(genes, sample1_name, sample2_name, comp_name):
    r1 = sample_results(sample1_name)
    r2 = sample_results(sample2_name)
    rc = sample_results(comp_name)
    gene_ids_in_all = set(r1).intersection(set(r2)).intersection(set(rc))

    diff_filename = os.path.join(comp_name, "diff.txt")
    field_names = ["gene_id", "chrom", "start", "end", "strand",
                   "num_transcripts", "num_paths", "num_parameters",
                   "sample1_num_reads", "sample2_num_reads", "joint_num_reads",
                   "max_delta_psi",
                   "sample1_ll", "sample2_ll", "diff_ll",
                   "null_ll", "lrt_d", "df", "p_value"]
    diff_writer = table.FileWriter(diff_filename, field_names)
    for gene in genes:
        gene_id = gene["gene_id"]
        if gene_id not in gene_ids_in_all:
            print >>sys.stderr, "Warning: Gene ID not in all samples:", gene_id
            continue
        max_delta_psi = compute_max_delta_psi(gene, sample1_name, sample2_name)
        result = dict(gene)
        result["max_delta_psi"] = max_delta_psi
        result["sample1_num_reads"] = int(r1[gene_id]["num_reads"])
        result["sample2_num_reads"] = int(r2[gene_id]["num_reads"])
        result["joint_num_reads"] = int(rc[gene_id]["num_reads"])
        result["sample1_ll"] = float(r1[gene_id]["log_likelihood"])
        result["sample2_ll"] = float(r2[gene_id]["log_likelihood"])
        result["diff_ll"] = result["sample1_ll"] + result["sample2_ll"]
        result["null_ll"] = float(rc[gene_id]["log_likelihood"])
        result["lrt_d"] = 2 * (result["diff_ll"] - result["null_ll"])
        result["df"] = int(gene["num_parameters"])
        result["p_value"] = chisquare_p(result["lrt_d"], result["df"])
        diff_writer.write(result)
    diff_writer.close()

def gene_sample_align_filenames(gene, sample):
    return [os.path.join(sample, gene["chrom"],
                         "%s.%d.align" % (gene["gene_id"], end))
            for end in [1,2]]
    
def concat_alignments(genes, sample1_name, sample2_name, comp_name):
    for gene in genes:
        sample_names = [sample1_name, sample2_name]
        in_filenames = zip(*[gene_sample_align_filenames(gene, sample)
                             for sample in sample_names])
        out_filenames = gene_sample_align_filenames(gene, comp_name)
        for in_end_filenames, out_end_filename in \
                zip(in_filenames, out_filenames):
            out_file = open(out_end_filename, 'w')
            for in_end_filename in in_end_filenames:
                if os.path.exists(in_end_filename):
                    in_file = open(in_end_filename)
                    out_file.write(in_file.read())
                    in_file.close()
            out_file.close()
        
def main():
    usage = "usage: %prog [options] reference_name sample1_name sample2_name comp_name"
    parser = optparse.OptionParser(usage=usage)
    parser.add_option("--frag-mean", dest="fragment_length_mean", 
                      help="Fragment length mean (default: 200)",
                      type="int", default=200, metavar="MEAN")
    parser.add_option("--frag-stddev", dest="fragment_length_stddev",
                      help="Fragment length standard deviation (default: 25)",
                      type="float", default=25.0, metavar="STDDEV")
    options, args = parser.parse_args()
    if len(args) != 4:
        parser.error("Invalid number of arguments")
    ref_name, sample1_name, sample2_name, comp_name = args

    try:
        num_cores = multiprocessing.cpu_count()
        print >>sys.stderr, "Using %d cores" % num_cores
        pool = multiprocessing.Pool(processes=num_cores)

        if not os.path.exists(comp_name):
            os.makedirs(comp_name)

        # Read reference index
        index_filename = os.path.join(ref_name, "index.txt")
        genes = list(table.FileReader(index_filename))
        gene_id_to_chrom_map = {}
        for rec in genes:
            gene_id_to_chrom_map[rec["gene_id"]] = rec["chrom"]

        # Make chromosome output dirs
        for chrom in set(gene_id_to_chrom_map.values()):
            sample_chrom_dir = os.path.join(comp_name, chrom)
            if not os.path.exists(sample_chrom_dir):
                os.makedirs(sample_chrom_dir)

        # Create alignment files for each PSG
        psg.log("Splitting alignment...")
        for sample in (sample1_name, sample2_name):
            for end in (1, 2):
                align_filename = os.path.join(sample,
                                              "alignment%d.bowtie.gz" % end)
                psg.split_alignment(align_filename,
                                    end,
                                    comp_name,
                                    gene_id_to_chrom_map,
                                    append=(sample == sample2_name))
        
        # Run PSG inferrer
        psg.infer_frequencies(genes,
                              comp_name,
                              ref_name,
                              options.fragment_length_mean,
                              options.fragment_length_stddev,
                              pool)
        
        # Perform differential splicing test
        psg.log("Performing differential splicing test")
        test_diff_splicing(genes, sample1_name, sample2_name, comp_name)
        
    except psg.Error, e:
        print >>sys.stderr, "Error: %s" % e.message
        sys.exit(1)

if __name__ == "__main__":
    main()
