#!/usr/bin/env python

import sys
import optparse
import collections
import os.path
import os
import subprocess
import multiprocessing
import itertools

import gff
import splicegraph
import fasta
import string
import psg
import table

def run_bowtie(reads_filename, ref_name, align_filename, num_cores,
               temp_dir=None, sort_buffer_size=None):
    bowtie_prefix = os.path.join(ref_name, "references")
    cmd = ("bowtie --best --strata -p %d -l 25 -a -m 200 %s %s 2> %s" %
           (num_cores,
            bowtie_prefix,
            reads_filename,
            align_filename + ".log"))
    sort_cmd = "sort -k3,3"
    if temp_dir is not None:
        sort_cmd += " -T " + temp_dir
    if sort_buffer_size is not None:
        sort_cmd += " -S " + sort_buffer_size
    cmd += " | %s | gzip -c > %s" % (sort_cmd, align_filename)
    psg.log(cmd)
    psg.run_command(cmd)

def main():
    usage = "usage: %prog [options] upstream_read_file downstream_read_file reference_name sample_name"
    parser = optparse.OptionParser(usage=usage)
    parser.add_option("--frag-mean", dest="fragment_length_mean", 
                      help="Fragment length mean (default: 200)",
                      type="int", default=200, metavar="MEAN")
    parser.add_option("--frag-stddev", dest="fragment_length_stddev",
                      help="Fragment length standard deviation (default: 25)",
                      type="float", default=25.0, metavar="STDDEV")
    parser.add_option("--pseudocount", dest="beta",
                      help="Prior pseudocount for each edge.  Set this to "
                      "zero for maximum likelihood estimates.  Set to a "
                      "positive value to obtain smoother estimates when the "
                      "number of reads is small (default: 1)",
                      type="float", default=1, metavar="COUNT")
    parser.add_option("--tempdir", dest="tempdir",
                      help="Temporary directory used for sorting alignments "
                      "(default: use default for 'sort' command)",
                      type="str", default=None, metavar="DIR")
    parser.add_option("--sort-buffer-size", dest="sort_buffer_size",
                      help="Buffer size in memory for sorting alignments "
                      "(default: use default for 'sort' command)",
                      default=None, metavar="SIZE")
    options, args = parser.parse_args()
    if len(args) != 4:
        parser.error("Invalid number of arguments")
    end1_filename, end2_filename, ref_name, sample_name = args

    try:
        num_cores = multiprocessing.cpu_count()
        print >>sys.stderr, "Using %d cores" % num_cores
        pool = multiprocessing.Pool(processes=num_cores)

        if not os.path.exists(sample_name):
            os.makedirs(sample_name)

        # Read reference index
        index_filename = os.path.join(ref_name, "index.txt")
        genes = list(table.FileReader(index_filename))
        gene_id_to_chrom_map = {}
        for rec in genes:
            gene_id_to_chrom_map[rec["gene_id"]] = rec["chrom"]

        # Make chromosome output dirs
        for chrom in set(gene_id_to_chrom_map.values()):
            sample_chrom_dir = os.path.join(sample_name, chrom)
            if not os.path.exists(sample_chrom_dir):
                os.makedirs(sample_chrom_dir)

        # Run Bowtie
        end1_align_filename = os.path.join(sample_name, "alignment1.bowtie.gz")
        end2_align_filename = os.path.join(sample_name, "alignment2.bowtie.gz")
        run_bowtie(end1_filename, ref_name, end1_align_filename, num_cores,
                   options.tempdir, options.sort_buffer_size)
        run_bowtie(end2_filename, ref_name, end2_align_filename, num_cores,
                   options.tempdir, options.sort_buffer_size)

        # Create alignment files for each PSG
        psg.log("Splitting alignment...")
        psg.split_alignment(end1_align_filename, 1, sample_name, gene_id_to_chrom_map)
        psg.split_alignment(end2_align_filename, 2, sample_name, gene_id_to_chrom_map)

        # Run PSG inferrer
        psg.infer_frequencies(genes,
                              sample_name,
                              ref_name,
                              options.fragment_length_mean,
                              options.fragment_length_stddev,
                              options.beta,
                              pool)
        
    except psg.Error, e:
        print >>sys.stderr, "Error: %s" % e.message
        sys.exit(1)

if __name__ == "__main__":
    main()
