class Record:
    def __init__(self,
                 qname="",
                 flag=0,
                 rname="",
                 pos=0,
                 mapq=0,
                 cigar="",
                 rnext="",
                 pnext=0,
                 tlen=0,
                 seq="",
                 qual="",
                 tag_tokens=None):
        self.qname = qname
        self.flag = flag
        self.rname = rname
        self.pos = pos
        self.mapq = mapq
        self.cigar = cigar
        self.rnext = rnext
        self.pnext = pnext
        self.tlen = tlen
        self.seq = seq
        self.qual = qual

        self.tag_tokens = tag_tokens if tag_tokens is not None else []
        self.tags = {}
        self.tags_parsed = False

    def query_name(self): return self.qname
    def flag(self): return self.flag
    def reference_name(self): return self.rname
    def reference_position(self): return self.pos
    def mapping_quality(self): return self.mapq
    def cigar_string(self): return self.cigar
    def next_reference_name(self): return self.rname
    def next_reference_position(self): return self.pnext
    def template_length(self): return self.tlen
    def fragment_sequence(self): return self.seq
    def fragment_qualities(self): return self.qual

    def template_has_multiple_fragments(self):
        return bool(self.flag & 0x1)
    def all_fragments_aligned(self):
        return bool(self.flag & 0x2)
    def is_unmapped(self):
        return bool(self.flag & 0x4)
    def next_is_unmapped(self):
        return bool(self.flag & 0x8)
    def is_reverse_complemented(self):
        return bool(self.flag & 0x10)
    def next_is_reverse_complemented(self):
        return bool(self.flag & 0x20)
    def is_first_fragment(self):
        return bool(self.flag & 0x40)
    def is_last_fragment(self):
        return bool(self.flag & 0x80)
    def is_secondary_alignment(self):
        return bool(self.flag & 0x100)
    def fails_quality_controls(self):
        return bool(self.flag & 0x200)
    def is_duplicate(self):
        return bool(self.flag & 0x400)
    
    def tag_value(self, tag_name):
        if not self.tags_parsed:
            self._parse_tags()
        return self.tags.get(tag_name, None)

    def _parse_tags(self):
        self.tags = tag_dict(self.tag_tokens)
        self.tags_parsed = True
        
array_type_map = {"c": ("b", int),
                  "C": ("B", int),
                  "s": ("h", int),
                  "S": ("H", int),
                  "i": ("l", long),
                  "I": ("L", long),
                  "f": ("f", float)}

def array_from_string(s):
    sam_array_type = s[1]
    typecode, array_type = array_type_map[sam_array_type]
    array_values = map(array_type, s[2:].split(","))
    return array.array(typecode, values)

def bytearray_from_hex_string(s):
    return bytearray(s.decode("hex"))

tag_value = {"A": str,
             "i": int,
             "f": float,
             "Z": str,
             "H": bytearray_from_hex_string,
             "B": array_from_string}

def tag_pair(tag):
    name, tag_type, value = tag.split(":", 2)
    return (name, tag_value[tag_type](value))
    
def tag_dict(tags):
    return dict(map(tag_pair, tags))

def none_if_unset(value, unset_value):
    return value if value != unset_value else None

def Reader(stream):
    for line in stream:
        if line.startswith("@"): continue
        fields = line[:-1].split('\t')
        yield Record(none_if_unset(fields[0], "*"),
                     int(fields[1]),
                     none_if_unset(fields[2], "*"),
                     none_if_unset(int(fields[3]), 0),
                     none_if_unset(int(fields[4]), 255),
                     none_if_unset(fields[5], "*"),
                     (fields[2] if fields[6] == "=" else
                      none_if_unset(fields[6], "*")),
                     none_if_unset(int(fields[7]), 0),
                     none_if_unset(int(fields[8]), 0),
                     none_if_unset(fields[9], "*"),
                     none_if_unset(fields[10], "*"),
                     fields[11:])
