#!/usr/bin/env python

import random
import operator
import sys
import math
import itertools

def dirichlet_sample(k, alpha=1.0):
    gamma_samples = [random.gammavariate(alpha, 1) for i in xrange(k)]
    sample_sum = sum(gamma_samples)
    return [sample / sample_sum for sample in gamma_samples]

def counts(lst):
    c = {}
    for elt in lst:
        c[elt] = c.get(elt, 0) + 1
    return c

def transform_segment(seg, start, end, strand):
    if strand == '+':
        return (seg[0] - start, seg[1] - start)
    else:
        return (end - seg[1], end - seg[0])

def transform_segment_list(seg_list, start, end, strand):
    transformed = [transform_segment(seg, start, end, strand)
                   for seg in seg_list]
    if strand == '-':
        transformed.reverse()
    length = end - start
    return [(0, 0)] + transformed + [(length, length)]

def interval_list_pairs(interval_list, interval_map):
    return [(interval_map[i1], interval_map[i2]) 
            for i1, i2 in zip(interval_list, interval_list[1:])]

def splice_points(segment_list, start, end, strand):
    points = [start] + joined_lists(segment_list) + [end]
    if strand == '+':
        normalized = [p - start for p in points]
    else:
        normalized = [end - p for p in reversed(points)]
    return normalized

def kth_order_segments(points, k):
    segs = [tuple(points[max(0, i - k - 1): i]) for i in xrange(3, len(points), 2)]
    return [(0, 0)] + segs + [(points[-1], points[-1])]

def segment_interval(segment):
    return segment[-2:]

def consecutive_pairs(lst):
    return zip(lst, lst[1:])

def zeroth_order_intervals_and_paths(splice_point_lists):
    points = set()
    for splice_point_list in splice_point_lists:
        points.update(splice_point_list)
    sorted_points = sorted(points)

    point_index = dict([(point, i) for i, point in enumerate(sorted_points)])

    exon_begin_points = set()
    for splice_point_list in splice_point_lists:
        exon_spans = zip(splice_point_list[1::2], splice_point_list[2::2])
        for start, end in exon_spans:
            begins = sorted_points[point_index[start]: point_index[end]]
            exon_begin_points.update(begins)
    
    intervals = [(0, 0)]
    num_intervals_up_to_point = {}
    for start, end in zip(sorted_points, sorted_points[1:]):
        num_intervals_up_to_point[start] = len(intervals)
        if start in exon_begin_points:
            intervals.append((start, end))
    last_point = sorted_points[-1]
    num_intervals_up_to_point[last_point] = len(intervals)
    intervals.append((last_point, last_point))

    paths = []
    for splice_point_list in splice_point_lists:
        path = [0]
        exon_spans = zip(splice_point_list[1::2], splice_point_list[2::2])
        for start, end in exon_spans:
            exon_intervals = range(num_intervals_up_to_point[start],
                                   num_intervals_up_to_point[end])
            path.extend(exon_intervals)
        path.append(len(intervals) - 1)
        paths.append(path)

    return intervals, paths

def kth_order_intervals_and_paths(splice_point_lists, k):
    assert(k > 0)
    kth_order_segment_lists = [kth_order_segments(points, k)
                               for points in splice_point_lists]
    # Sort by segment interval
    segments = sorted(unique(joined_lists(kth_order_segment_lists)),
                      key=lambda segment: (segment_interval(segment), segment))
    segment_map = dict([(seg, index) for index, seg in enumerate(segments)])
    intervals = map(segment_interval, segments)
    paths = [[segment_map[s] for s in seg_list]
             for seg_list in kth_order_segment_lists]
    return intervals, paths

def kth_order_splice_graph(chrom, strand, segment_lists, db, k):
    start = min([segment_list[0][0] for segment_list in segment_lists])
    end = max([segment_list[-1][1] for segment_list in segment_lists])
    splice_point_lists = [splice_points(segment_list, start, end, strand)
                          for segment_list in segment_lists]
    if k == 0:
        ints_and_paths = zeroth_order_intervals_and_paths(splice_point_lists)
    else:
        ints_and_paths = kth_order_intervals_and_paths(splice_point_lists, k)
    intervals, paths = ints_and_paths

    pairs = joined_lists(map(consecutive_pairs, paths))
    pair_counts = counts(pairs)
    edges = [Edge(pair[0], pair[1], float(pair_counts[pair]))
             for pair in sorted(unique(pairs))]
    sequence = db.sequence(chrom, start, end, strand)
    segments = [sequence[i[0]: i[1]] for i in intervals]
    splice_graph = SpliceGraph(segments, edges)
    splice_graph.normalize()
    return (splice_graph, intervals, paths)

def make_splice_graph(chrom, strand, segment_lists, db):
    start = min([segment_list[0][0] for segment_list in segment_lists])
    end = max([segment_list[-1][1] for segment_list in segment_lists])

    transformed = [transform_segment_list(seg_list, start, end, strand)
                   for seg_list in segment_lists]
    intervals = unique(joined_lists(transformed))
    intervals.sort()
    interval_map = dict([(i, index) for index, i in enumerate(intervals)])

    pairs = joined_lists([interval_list_pairs(int_list, interval_map)
                          for int_list in transformed])
    pair_counts = counts(pairs)

    edges = [Edge(pair[0], pair[1], float(pair_counts[pair]))
             for pair in sorted(unique(pairs))]
    sequence = db.sequence(chrom, start, end, strand)
    segments = [sequence[i[0]: i[1]] for i in intervals]
    splice_graph = SpliceGraph(segments, edges)
    splice_graph.normalize()
    return splice_graph

def unfactored_splice_graph(psg):
    segments = [""]
    edges = []
    all_paths = psg.allPaths()
    num_segments = sum(map(len, all_paths)) - 2 * (len(all_paths) - 1)
    for path in all_paths:
        new_path = [0]
        for old_index in path[1:-1]:
            new_path.append(len(segments))
            segments.append(psg.segments[old_index])
        new_path.append(num_segments - 1)
        for start, end in zip(new_path, new_path[1:]):
            edges.append(Edge(start, end, prob=1.0))
    segments.append(psg.segments[-1])
    edges.sort(key=lambda edge: (edge.source, edge.target))
    return SpliceGraph(segments, edges)

def log(s):
    print >>sys.stderr, s

def list_of_lists(length):
    return [[] for i in range(length)]

def matrix(rows, cols, val=None, func=None):
    if func:
        return [[func() for col in xrange(cols)] for row in xrange(rows)]
    else:
        return [[val for col in xrange(cols)] for row in xrange(rows)]

def product(lst):
    return reduce(operator.mul, lst, 1.0)

def unique(lst):
    return list(set(lst))

def joined_lists(lists):
    joined = []
    for lst in lists:
        joined.extend(lst)
    return joined

def pct_changes(v1, v2):
    return [float(abs(e2 - e1)) / e1 for e1, e2 in zip(v1, v2)]

class Edge:
    def __init__(self, source, target, prob=0.0):
        self.source = source
        self.target = target
        self.prob = prob

class SpliceGraph:
    def __init__(self, segments, edges):
        self.segments = segments
        self.edges = edges

        # Construct adjacency lists for easy access
        self.in_edges = list_of_lists(len(self.segments))
        self.out_edges = list_of_lists(len(self.segments))
        for edge in self.edges:
            self.in_edges[edge.target].append(edge)
            self.out_edges[edge.source].append(edge)

        # Construct ancestry lists
        self.ancestors = list_of_lists(len(self.segments))
        for i in xrange(len(self.segments)):
            parents = [e.source for e in self.in_edges[i]]
            seg_ancestors = joined_lists([parents] +
                                         [self.ancestors[j] for j in parents])
            self.ancestors[i] = sorted(unique(seg_ancestors))

        self.normalize()
        self.makeTransitionMatrix()

    def getNumIsoforms(self):
        num_splicings = [0] * len(self.segments)
        num_splicings[0] = 1
        for i in xrange(1, len(self.segments)):
            for edge in self.in_edges[i]:
                num_splicings[i] += num_splicings[edge.source]
        return num_splicings[-1]

    def makeTransitionMatrix(self):
        self.transitions = matrix(len(self.segments), len(self.segments), 0.0)
        for edge in self.edges:
            self.transitions[edge.source][edge.target] = edge.prob

    def entropy(self, base=math.e):
        e = 0
        segment_probs = [1] * len(self.segments)
        for i in xrange(len(self.segments)):
            prob_segment_used = segment_probs[i]
            for edge in self.out_edges[i]:
                prob_edge_used = prob_segment_used * edge.prob
                if prob_edge_used > 0:
                    e += prob_edge_used * math.log(edge.prob)
                segment_probs[edge.target] += prob_edge_used
        e /= math.log(base)
        return e

    def relative_entropy(self, q, base=math.e):
        try:
            re = 0
            segment_probs = [1] * len(self.segments)
            for i in xrange(len(self.segments)):
                prob_segment_used = segment_probs[i]
                for edge_p, edge_q in zip(self.out_edges[i], q.out_edge[i]):
                    prob_edge_used = prob_segment_used * edge_p.prob
                    if prob_edge_used > 0:
                        re += prob_edge_used * math.log(edge_p.prob /
                                                        edge_q.prob)
                    segment_probs[edge.target] += prob_edge_used
            re /= math.log(base)
            return re
        except ZeroDivisionError:
            return None
    
    def getSegmentSequence(self, i):
        return self.segments[i]
    
    def samplePath(self):
        path = [0]
        while path[-1] != len(self.segments) - 1:
            out = self.out_edges[path[-1]]
            probs = [edge.prob for edge in out]
            segs = [edge.target for edge in out]
            path.append(segs[dist_sample(probs)])
        return path

    def allPaths(self):
        prefix_paths = [[] for i in xrange(len(self.segments))]
        prefix_paths[0] = [[0]]
        for i in xrange(1, len(self.segments)):
            for edge in self.in_edges[i]:
                prefix_paths[i].extend([p + [i] for p in prefix_paths[edge.source]])
        assert(len(prefix_paths[-1]) == self.getNumIsoforms())
        return prefix_paths[-1]
    
    def sampleTranscript(self):
        return ''.join([self.getSegmentSequence(i) for i in self.samplePath()])

    def sampleReads(self, read_len, num_reads):
        return [self.sampleRead(read_len) for i in xrange(num_reads)]

    def sampleRead(self, read_len):
        transcript = self.sampleTranscript()
        assert(len(transcript) >= read_len)
        start = random.randint(0, len(transcript) - read_len)
        return transcript[start: start + read_len]

    def hasEdge(self, i, j):
        return j in [edge.target for edge in self.out_edges[i]]

    def normalize(self):
        for i in range(len(self.segments)):
            self.normalizeOutEdges(i)
        self.makeTransitionMatrix()

    def randomizeProbs(self):
        for i in xrange(len(self.segments)):
            edges = self.out_edges[i]
            probs = dirichlet_sample(len(edges))
            for edge, prob in zip(edges, probs):
                edge.prob = prob
        
    def resetProbs(self):
        for i in range(len(self.segments)):
            self.resetOutEdges(i)
        self.makeTransitionMatrix()

    def resetOutEdges(self, i):
        seg_out_edges = self.out_edges[i]
        for edge in seg_out_edges:
            edge.prob = 1.0 / len(seg_out_edges)

    def normalizeOutEdges(self, i):
        seg_out_edges = self.out_edges[i]
        total = sum([edge.prob for edge in seg_out_edges])
        for edge in seg_out_edges:
            if total > 0:
                edge.prob /= total
            else:
                edge.prob = 1.0 / len(seg_out_edges)

    def averageOutDegree(self):
        return (sum([len(out) for out in self.out_edges]) / 
                float(len(self.segments) - 1))

    def readSubPaths(self, read):
        sub_paths = []
        # Check for read entirely in segment
        for i in range(len(self.segments)):
            if self.getSegmentSequence(i).find(read) != -1:
                sub_paths.append([i])
        # Check for junction spanning reads (right now assuming that
        # reads can not span more than one junction)
        for i in range(1, len(self.segments)):
            for j in [edge.target for edge in self.out_edges[i]]:
                junction = (self.getSegmentSequence(i)[-(len(read) - 1):] +
                            self.getSegmentSequence(j)[:(len(read) - 1)])
                if junction.find(read) != -1:
                    sub_paths.append([i, j])
        return sub_paths

    def subpathProb(self, sub_path):
        return product([self.transitions[i][j]
                        for i, j in zip(sub_path, sub_path[1:])])

    def subpathLen(self, sub_path):
        return sum([self.segments[i].length() for i in sub_path])

    def probReadFromSubpath(self, sub_path, read_len):
        l_subpath = self.subpathLen(sub_path)
        prob = 0
        for l1, p1 in self.f[0][sub_path[0]].iteritems():
            for l2, p2 in self.f[sub_path[-1]][-1].iteritems():
                prob += p1 * p2 * (1.0 / (l1 + l2 + l_subpath - read_len + 1))
        prob *= self.subpathProb(sub_path)
        return prob

    def probReadFromSubpathPair(self, sp1, sp2, read_len):
        if sp1 == sp2 or sp2 == sp1[:len(sp2)] or sp2 == sp1[-len(sp2):]:
            return self.probReadFromSubpath(sp1, read_len)
        if sp1[0] > sp2[0]:
            sp1, sp2 = sp2, sp1
        l_subpath = self.subpathLen(sp1) + self.subpathLen(sp2)
        prob = 0
        for l1, p1 in self.f[0][sp1[0]].iteritems():
            for l2, p2 in self.f[sp1[-1]][sp2[0]].iteritems():
                for l3, p3 in self.f[sp2[-1]][-1].iteritems():
                    prob += p1 * p2 * p3 * (1.0 / (l1 + l2 + l3 + l_subpath - read_len + 1))
        prob *= self.subpathProb(sp1) * self.subpathProb(sp2)
        return prob

    def logLikelihood(self, read_sub_paths, read_len):
        self.computeLengthProbs()
        log_likelihood = 0.0
        for sub_paths in read_sub_paths:
            p = sum([self.probReadFromSubpath(sp, read_len) for sp in sub_paths])
            log_likelihood += math.log(p)
        return log_likelihood

    def eStep(self, read_sub_paths, read_len):
        self.computeLengthProbs()
        log_likelihood = 0.0
        edge_counts = [0] * len(self.edges)
        for sub_paths in read_sub_paths:
            read_prob = sum([self.probReadFromSubpath(sp, read_len) for sp in sub_paths])
            log_likelihood += math.log(read_prob)
            for i in xrange(len(edge_counts)):
                edge = self.edges[i]
                ep = [edge.source, edge.target]
                n = sum([self.probReadFromSubpathPair(ep, sp, read_len) for sp in sub_paths])
                e = n / read_prob
                edge_counts[i] += e
        return log_likelihood, edge_counts

    def mStep(self, edge_counts):
        # Count number of times each segment is used
        seg_counts = [0] * len(self.segments)
        for edge, count in zip(self.edges, edge_counts):
            seg_counts[edge.source] += count
        # Compute probs from edge counts and seg counts
        for edge, count in zip(self.edges, edge_counts):
            edge.prob = count / seg_counts[edge.source]
        self.makeTransitionMatrix()
    
    def getEdgeProbs(self):
        return [edge.prob for edge in self.edges]

    def setEdgeProbs(self, probs):
        for prob, edge in zip(probs, self.edges):
            edge.prob = prob
        self.makeTransitionMatrix()

    def learnEM(self, reads,
                min_prob_change=0.002, min_ll_change=0.001, max_iters=1000,
                quiet=False):
        read_len = len(reads[0])
        read_sub_paths = map(self.readSubPaths, reads)
        ll = None
        probs = self.getEdgeProbs()
        iter = 0
        while True:
            prev_ll = ll
            prev_probs = probs
            ll, edge_counts = self.eStep(read_sub_paths, read_len)
            self.mStep(edge_counts)
            probs = self.getEdgeProbs()
            if not quiet:
                log("%d %.3f " % (iter, ll) + 
                    ' '.join(["%.3f" % p for p in prev_probs]))
            iter += 1
            if (max(pct_changes(prev_probs, probs)) < min_prob_change or
                (prev_ll and ll - prev_ll < min_ll_change) or
                iter == max_iters):
                break

    def computeLengthProbs(self):
        lengths = [s.length() for s in self.segments]
        num_segments = len(self.segments)
        f = matrix(num_segments, num_segments, func=dict)
        for j in xrange(num_segments):
            f[j][j] = {0: 1.0}
            for edge in self.in_edges[j]:
                k = edge.source
                p_kj = edge.prob
                l_k = lengths[k]
                assert(k < j and p_kj > 0)
                f[k][j] = {0: p_kj}
                for i in self.ancestors[k]:
                    for l_ik, p_ik in f[i][k].iteritems():
                        l = l_ik + l_k
                        f[i][j][l] = f[i][j].get(l, 0) + p_ik * p_kj
        self.f = f

    def getExpectedLength(self):
        self.computeLengthProbs()
        return sum([length * prob for length, prob in self.f[0][-1].items()])

    def getNumFreeParameters(self):
        return sum([len(out_edges) - 1 for out_edges in self.out_edges[:-1]])

    def printLengthProbs(self):
        for i in xrange(len(self.segments)):
            for j in xrange(len(self.segments)):
                for l, p in sorted(self.f[i][j].items()):
                    print "%d\t%d\t%d\t%.3f" % (i, j, l, p)

def dist_sample(dist):
    x = random.random()
    i = 0
    while x > dist[i]:
        x -= dist[i]
        i += 1
    return i

def is_data_line(line):
    return line.strip() and not line.startswith("#")

def lines_to_fields(lines):
    return [line.split() for line in lines]

def read_segments(lines):
    labeled_segments = [(int(f[0]), f[1].strip() if len(f) > 1 else "") for f in lines_to_fields(lines)]
    labeled_segments.sort()
    assert(labeled_segments[0][0] == 0)
    assert(labeled_segments[-1][0] == (len(labeled_segments) - 1))
    return [s[1] for s in labeled_segments]

def read_edges(lines):
    return [Edge(int(f[0]), int(f[1]), float(f[2])) 
            for f in lines_to_fields(lines)]

def read_sections(stream, readers):
    sections = {}
    lines = stream.readlines()
    section_start = None
    line_num = 0
    while line_num <= len(lines):
        if line_num == len(lines) or lines[line_num].startswith('['):
            if section_start is not None:
                section = lines[section_start].rstrip()[1:-1]
                section_lines = filter(is_data_line,
                                       lines[section_start + 1: line_num])
                sections[section] = readers[section](section_lines)
            section_start = line_num
        line_num += 1
    return sections

def read_splice_graph(stream):
    section_readers = {'segments': read_segments,
                       'edges': read_edges}
    sections = read_sections(stream, section_readers)
    try:
        return SpliceGraph(sections['segments'], 
                           sections['edges'])
    except KeyError, e:
        raise Exception("Required section not found in splice graph file:" 
                        + e.message)

def write_segments(stream, segments):
    print >>stream, '[segments]'
    for i, segment in enumerate(segments):
        print >>stream, "%d\t%s" % (i, segment)

def write_edges(stream, edges):
    print >>stream, '[edges]'
    for edge in edges:
        print >>stream, "%d\t%d\t%f" % (edge.source, edge.target, edge.prob)

def write_splice_graph(stream, splice_graph):
    write_segments(stream, splice_graph.segments)
    write_edges(stream, splice_graph.edges)
